# Copyright 2019 The ASReview Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from asreview.feature_extraction.doc2vec import Doc2Vec
from asreview.feature_extraction.tfidf import Tfidf
from asreview.feature_extraction.sbert import SBERT
from asreview.feature_extraction.embedding_idf import EmbeddingIdf
from asreview.feature_extraction.embedding_lstm import EmbeddingLSTM


def get_feature_class(method):
    """Get class of feature extraction from string.

    Arguments
    ---------
    method: str
        Name of the feature model, e.g. 'doc2vec', 'tfidf' or 'embedding-lstm'.

    Returns
    -------
    BaseFeatureExtraction:
        Class corresponding to the method.
    """
    models = {
        "doc2vec": Doc2Vec,
        "tfidf": Tfidf,
        "sbert": SBERT,
        "embedding-idf": EmbeddingIdf,
        "embedding-lstm": EmbeddingLSTM,
    }
    try:
        return models[method]
    except KeyError:
        raise ValueError(f"Unsupervised method '{method}' does not exist.")


def get_feature_model(method, *args, **kwargs):
    """Get an instance of a feature extraction model from a string.

    Arguments
    ---------
    method: str
        Name of the feature extraction model.
    *args:
        Arguments for the feature extraction model.
    **kwargs:
        Keyword arguments for thefeature extraction  model.
    """
    model_class = get_feature_class(method)
    return model_class(*args, **kwargs)



# Copyright 2019 The ASReview Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC, abstractmethod

import numpy as np


class BaseLogger(ABC):
    def __init__(self, log_fp, read_only=False):
        self.log_fp = log_fp
        self.read_only = read_only
        self.settings = None
        self.restore(log_fp)

    def __enter__(self):
        return self

    def __exit__(self, *_, **__):
        self.close()

    def __str__(self):
        return str(self.to_dict())

    @abstractmethod
    def set_labels(self, y):
        """Add/set labels to logger

        If the labels do not exist, add it to the logger.

        Arguments
        ---------
        y: np.array
            One dimensional integer numpy array with inclusion labels.
        """
        raise NotImplementedError

    @abstractmethod
    def set_final_labels(self, y):
        """Add/set final labels to logger.

        If final_labels does not exist yet, add it.

        Arguments
        ---------
        y: np.array
            One dimensional integer numpy array with final inclusion labels.
        """
        raise NotImplementedError

    @abstractmethod
    def add_settings(self, settings):
        """Add settings to the logger.

        Arguments
        ---------
        settings: ASReviewSettings
            Settings object to add to the logger.
        """
        raise NotImplementedError

    @abstractmethod
    def add_classification(self, idx, labels, methods, query_i):
        """Add training indices and their labels.

        Arguments
        ---------
        indices: list, np.array
            A list of indices used for training.
        labels: list
            A list of labels corresponding with the training indices.
        i: int
            The query number.
        """
        raise NotImplementedError

    @abstractmethod
    def add_proba(self, pool_idx, train_idx, proba, query_i):
        """Add inverse pool indices and their labels.

        Arguments
        ---------
        indices: list, np.array
            A list of indices used for unlabeled pool.
        pred: np.array
            Array of prediction probabilities for unlabeled pool.
        i: int
            The query number.
        """
        raise NotImplementedError

    def is_empty(self):
        """Check if logger has no results.

        Returns
        -------
        bool:
            True if empty.
        """
        return self.n_queries() == 0

    @abstractmethod
    def n_queries(self):
        """Number of queries saved in the logger.

        Returns
        -------
        int:
            Number of queries.
        """
        raise NotImplementedError

    @abstractmethod
    def get(self, variable, query_i=None, idx=None):
        """Get data from the logger object.

        This is universal accessor method of the Logger classes. It can be used
        to get a variable from one specific query. In theory, it should get the
        whole data set if query_i=None, but this is not currently implemented
        in any of the Loggers.

        Arguments
        ---------
        variable: str
            Name of the variable/data to get. Options are:
            label_idx, inclusions, label_methods, labels, final_labels, proba
            , train_idx, pool_idx.
        query_i: int
            Query number, should be between 0 and self.n_queries().
        idx: int, np.array, list
            Indices to get in the returned array.
        """
        raise NotImplementedError

    @abstractmethod
    def delete_last_query(self):
        """Delete the last query from the logger object."""
        raise NotImplementedError

    def review_state(self):
        """Get variables for reviewer to continue review.

        Returns
        -------
        np.array:
            Current labels of dataset.
        np.array:
            Current training indices.
        dict:
            Dictionary containing the sources of the labels.
        query_i:
            Currenty query number (starting from 0).
        """
        labels = self.get("labels")

        train_idx = []
        query_src = {}
        for query_i in range(self.n_queries()):
            try:
                label_idx = self.get("label_idx", query_i)
                labelled = self.get("inclusions", query_i)
                label_methods = self.get("label_methods", query_i)
            except (KeyError, IndexError):
                continue

            for i, meth in enumerate(label_methods):
                if meth not in query_src:
                    query_src[meth] = []
                query_src[meth].append(label_idx[i])
                labels[label_idx[i]] = labelled[i]
            train_idx.extend(label_idx)

        if query_i > 0:
            n_queries = self.n_queries()
            last_inclusions = None
            try:
                last_inclusions = self.get("inclusions", n_queries-1)
            except KeyError:
                last_inclusions = None
            if last_inclusions is None or len(last_inclusions) == 0:
                query_i -= 1
                self.delete_last_query()

        train_idx = np.array(train_idx, dtype=np.int)
        return labels, train_idx, query_src, query_i

    @abstractmethod
    def initialize_structure(self):
        """Create empty internal structure for logger"""
        raise NotImplementedError

    @abstractmethod
    def close(self):
        """Close the files opened by the logger.

        Also sets the end time if not in read-only mode.
        """
        raise NotImplementedError

    @abstractmethod
    def save(self):
        """Save logs to file.

        Arguments
        ---------
        fp: str
            The file path to export the results to.

        """
        raise NotImplementedError

    @abstractmethod
    def restore(self, fp):
        """Restore or create logger from a log file.

        If the log file doesn't exist, creates and empty state that is ready
        for storage.

        Arguments
        ---------
        fp: str
            Path to file to restore/create.
        """
        raise NotImplementedError

    def to_dict(self):
        """Convert logger to dictionary.

        Returns
        -------
        dict:
            Dictionary with all relevant variables.
        """
        log_dict = {}
        log_dict["settings"] = vars(self.settings)

        global_datasets = ["labels", "final_labels"]
        for dataset in global_datasets:
            try:
                log_dict[dataset] = self.get(dataset).tolist()
            except KeyError:
                pass

        query_datasets = [
            "label_methods", "label_idx", "inclusions", "proba", "pool_idx",
            "train_idx"]
        log_dict["results"] = []
        for query_i in range(self.n_queries()):
            log_dict["results"].append({})
            for dataset in query_datasets:
                try:
                    log_dict["results"][query_i][dataset] = self.get(
                        dataset, query_i).tolist()
                except (KeyError, IndexError):
                    pass
        return log_dict



import logging

from asreview.ascii import welcome_message

from asreview.entry_points.oracle import _oracle_parser
from asreview.config import DEFAULT_N_PRIOR_EXCLUDED
from asreview.config import DEFAULT_N_PRIOR_INCLUDED
from asreview.entry_points.base import BaseEntryPoint
from asreview.review import review_simulate


class SimulateEntryPoint(BaseEntryPoint):
    description = "Simulate the performance of ASReview."

    def execute(self, argv):
        parser = _simulate_parser()
        args = parser.parse_args(argv)

        args_dict = vars(args)
        path = args_dict.pop("dataset")

        verbose = args_dict.get("verbose", 0)
        if verbose == 0:
            logging.getLogger().setLevel(logging.WARNING)
        elif verbose == 1:
            logging.getLogger().setLevel(logging.INFO)
        elif verbose >= 2:
            logging.getLogger().setLevel(logging.DEBUG)

        print(welcome_message("simulate"))
        review_simulate(path, **args_dict)


DESCRIPTION_SIMULATE = """
Automated Systematic Review (ASReview) for simulation runs.

The simulation modus is used to measure the performance of our
software on existing systematic reviews. The software shows how many
papers you could have potentially skipped during the systematic
review."""


def _simulate_parser():
    parser = _oracle_parser(prog="simulate", description=DESCRIPTION_SIMULATE)
    # Initial data (prior knowledge)
    parser.add_argument(
        "--n_prior_included",
        default=DEFAULT_N_PRIOR_INCLUDED,
        type=int,
        help="Sample n prior included papers. "
             "Only used when --prior_included is not given. "
             f"Default {DEFAULT_N_PRIOR_INCLUDED}")

    parser.add_argument(
        "--n_prior_excluded",
        default=DEFAULT_N_PRIOR_EXCLUDED,
        type=int,
        help="Sample n prior excluded papers. "
             "Only used when --prior_excluded is not given. "
             f"Default {DEFAULT_N_PRIOR_EXCLUDED}")
    return parser



import argparse
from argparse import RawTextHelpFormatter
import logging

from asreview.ascii import welcome_message

from asreview.config import DEFAULT_MODEL
from asreview.config import DEFAULT_QUERY_STRATEGY
from asreview.config import DEFAULT_BALANCE_STRATEGY
from asreview.config import DEFAULT_N_INSTANCES
from asreview.entry_points.base import BaseEntryPoint
from asreview.review import review_oracle


class OracleEntryPoint(BaseEntryPoint):
    description = "Perform a systematic review on the command line."

    def execute(self, argv):
        parser = _oracle_parser()
        args = parser.parse_args(argv)

        args_dict = vars(args)
        path = args_dict.pop("dataset")

        verbose = args_dict.get("verbose", 0)
        if verbose == 0:
            logging.getLogger().setLevel(logging.WARNING)
        elif verbose == 1:
            logging.getLogger().setLevel(logging.INFO)
        elif verbose >= 2:
            logging.getLogger().setLevel(logging.DEBUG)

        print(welcome_message("oracle"))
        review_oracle(path, **args_dict)


DESCRIPTION_ORACLE = """
Automated Systematic Review (ASReview) with interaction with oracle.

The oracle modus is used to perform a systematic review with
interaction by the reviewer (the ‘oracle’ in literature on active
learning). The software presents papers to the reviewer, whereafter
the reviewer classifies them."""


def _oracle_parser(prog="oracle", description=DESCRIPTION_ORACLE):
    """Argument parser for oracle and simulate.

    Parameters
    ----------
    mode : str
        The mode to run ASReview. Options:
        'simulate' and 'oracle'.
    prog : str
        The program name. For example 'asreview'.

    Returns
    -------
    argparse.ArgumentParser
        Configured argparser.
    """

    # parse arguments if available
    parser = argparse.ArgumentParser(
        prog=prog,
        description=description,
        formatter_class=RawTextHelpFormatter
    )
    # File path to the data.
    parser.add_argument(
        "dataset",
        type=str,
        metavar="X",
        help="File path to the dataset or one of the built-in datasets."
    )
    # Active learning parameters
    parser.add_argument(
        "-m", "--model",
        type=str,
        default=DEFAULT_MODEL,
        help=f"The prediction model for Active Learning. "
             f"Default '{DEFAULT_MODEL}'.")  #noqa
    parser.add_argument(
        "-q", "--query_strategy",
        type=str,
        default=DEFAULT_QUERY_STRATEGY,
        help=f"The query strategy for Active Learning. "
             f"Default '{DEFAULT_QUERY_STRATEGY}'.")  #noqa
    parser.add_argument(
        "-b", "--balance_strategy",
        type=str,
        default=DEFAULT_BALANCE_STRATEGY,
        help="Data rebalancing strategy mainly for RNN methods. Helps against"
             " imbalanced dataset with few inclusions and many exclusions. "
             f"Default '{DEFAULT_BALANCE_STRATEGY}'")
    parser.add_argument(
        "--n_instances",
        default=DEFAULT_N_INSTANCES,
        type=int,
        help="Number of papers queried each query."
             f"Default {DEFAULT_N_INSTANCES}.")
    parser.add_argument(
        "--n_queries",
        type=int,
        default=None,
        help="The number of queries. By default, the program "
             "stops after all documents are reviewed or is "
             "interrupted by the user."
    )
    parser.add_argument(
        "-n", "--n_papers",
        type=int,
        default=None,
        help="The number of papers to be reviewed. By default, "
             "the program stops after all documents are reviewed or is "
             "interrupted by the user."
    )
    parser.add_argument(
        "--embedding",
        type=str,
        default=None,
        dest='embedding_fp',
        help="File path of embedding matrix. Required for LSTM models."
    )
    # Configuration file with model/balance/query parameters.
    parser.add_argument(
        "--config_file",
        type=str,
        default=None,
        help="Configuration file with model parameters"
    )
    # Initial data (prior knowledge)
    parser.add_argument(
        "--prior_included",
        default=None,
        type=int,
        nargs="*",
        help="A list of included papers.")

    parser.add_argument(
        "--prior_excluded",
        default=None,
        type=int,
        nargs="*",
        help="A list of excluded papers. Optional.")

    parser.add_argument(
        "--extra_dataset",
        default=[],
        action='append',
        help="A dataset with labels to improve training. Can be used multiple"
             " times."
    )

    # logging and verbosity
    parser.add_argument(
        "--log_file", "-l",
        default=None,
        type=str,
        help="Location to store the log results."
    )
    parser.add_argument(
        "--save_model",
        default=None,
        type=str,
        dest='save_model_fp',
        help="Location to store the model and weights. "
             "Only works for Keras/RNN models. "
             "End file extension with '.json'."
    )
    parser.add_argument(
        "--verbose", "-v",
        default=0,
        type=int,
        help="Verbosity")

    return parser



# Copyright 2019 The ASReview Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from tensorflow.keras import optimizers
from tensorflow.keras.layers import Dense
from tensorflow.keras.layers import Embedding
from tensorflow.keras.layers import LSTM
from tensorflow.keras.models import Sequential
from tensorflow.keras.wrappers.scikit_learn import KerasClassifier

from asreview.models.base import BaseTrainModel
from asreview.utils import _set_class_weight


class LSTMBaseModel(BaseTrainModel):
    """ LSTM base class.

    LSTM model consisting of an embedding layer, one LSTM layer, and one
    dense layer.
    """
    name = "lstm-base"

    def __init__(self, embedding_matrix=None, backwards=True, dropout=0.4,
                 optimizer="rmsprop", lstm_out_width=20, learn_rate=1.0,
                 dense_width=128, verbose=0, batch_size=32, epochs=35,
                 shuffle=False, class_weight=30.0):
        """Initialize the LSTM base model.

        Arguments
        ---------
        embedding_matrix: np.array
            Embedding matrix to use with LSTM model.
        backwards: bool
            Whether to have a forward or backward LSTM.
        dropout: float
            Value in [0, 1.0) that gives the dropout and recurrent
            dropout rate for the LSTM model.
        optimizer: str
            Optimizer to use.
        lstm_out_width: int
            Output width of the LSTM.
        learn_rate: float
            Learn rate multiplier of default learning rate.
        dense_width: int
            Size of the dense layer of the model.
        verbose: int
            Verbosity.
        batch_size: int
            Size of the batch size for the LSTM model.
        epochs: int
            Number of epochs to train the LSTM model.
        shuffle: bool
            Whether to shuffle the data before starting to train.
        class_weight: float
            Class weight for the included papers.
        """
        super(LSTMBaseModel, self).__init__()
        self.embedding_matrix = embedding_matrix
        self.backwards = backwards
        self.dropout = dropout
        self.optimizer = optimizer
        self.lstm_out_width = lstm_out_width
        self.learn_rate = learn_rate
        self.dense_width = dense_width
        self.verbose = verbose
        self.batch_size = batch_size
        self.epochs = epochs
        self.shuffle = shuffle
        self.class_weight = _set_class_weight(class_weight)
        self._model = None
        self.sequence_length = None

    def fit(self, X, y):
        sequence_length = X.shape[1]
        if self._model is None or sequence_length != self.sequence_length:
            self.sequence_length = sequence_length
            keras_model = _create_lstm_base_model(
                embedding_matrix=self.embedding_matrix,
                backwards=self.backwards, dropout=self.dropout,
                optimizer=self.optimizer,
                max_sequence_length=self.sequence_length,
                lstm_out_width=self.lstm_out_width,
                dense_width=self.dense_width,
                learn_rate=self.learn_rate, verbose=self.verbose)
            print(keras_model)
            self._model = KerasClassifier(keras_model, verbose=self.verbose)

        self._model.fit(X, y, batch_size=self.batch_size, epochs=self.epochs,
                        shuffle=self.shuffle, class_weight=self.class_weight,
                        verbose=self.verbose)

    def full_hyper_space(self):
        from hyperopt import hp
        hyper_choices = {}
        hyper_space = {
            "mdl_dropout": hp.uniform("mdl_dropout", 0, 0.9),
            "mdl_lstm_out_width": hp.quniform("mdl_lstm_out_width", 1, 50, 1),
            "mdl_dense_width": hp.quniform("mdl_dense_width", 1, 200, 1),
            "mdl_learn_rate_mult": hp.lognormal("mdl_learn_rate_mult", 0, 1)
        }
        return hyper_space, hyper_choices


def _create_lstm_base_model(embedding_matrix,
                            backwards=True,
                            dropout=0.4,
                            optimizer='rmsprop',
                            max_sequence_length=1000,
                            lstm_out_width=20,
                            dense_width=128,
                            learn_rate=1.0,
                            verbose=1):
    """Return callable lstm model.
    Returns
    -------
    callable:
        A function that return the Keras Sklearn model when
        called.

    """
    def model_wrapper():
        model = Sequential()

        # add first embedding layer with pretrained wikipedia weights
        model.add(
            Embedding(
                embedding_matrix.shape[0],
                embedding_matrix.shape[1],
                weights=[embedding_matrix],
                input_length=max_sequence_length,
                trainable=False
            )
        )

        # add LSTM layer
        model.add(
            LSTM(
                lstm_out_width,
                input_shape=(max_sequence_length, ),
                go_backwards=backwards,
                dropout=dropout,
                recurrent_dropout=dropout,
            )
        )

        # add Dense layer with relu activation
        model.add(
            Dense(
                dense_width,
                activation='relu',
            )
        )

        # add Dense layer
        model.add(
            Dense(
                1,
                activation='sigmoid'
            )
        )

        optimizer_fn = _get_optimizer(optimizer, learn_rate)

        # Compile model
        model.compile(
            loss='binary_crossentropy',
            optimizer=optimizer_fn, metrics=['acc'])

        if verbose >= 1:
            model.summary(verbose=verbose)

        return model

    return model_wrapper


def _get_optimizer(optimizer, lr_mult=1.0):
    "Get optimizer with correct learning rate."
    if optimizer == "sgd":
        return optimizers.SGD(lr=0.01*lr_mult)
    elif optimizer == "rmsprop":
        return optimizers.RMSprop(lr=0.001*lr_mult)
    elif optimizer == "adagrad":
        return optimizers.Adagrad(lr=0.01*lr_mult)
    elif optimizer == "adam":
        return optimizers.Adam(lr=0.001*lr_mult)
    elif optimizer == "nadam":
        return optimizers.Nadam(lr=0.002*lr_mult)
    raise NotImplementedError
